﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.QuadraticAssignment;
using HeuristicLab.Random;

namespace HeuristicLab.Analysis.FitnessLandscape {
  [Item("Directed Walk (QAP-specific)", "")]
  [StorableClass]
  public class QAPDirectedWalk : CharacteristicCalculator {

    public enum WalkType { RandomRandom, RandomGlobal, RandomLocal, LocalLocal, LocalGlobal, LocalInverse }

    private const string NBHOOD_STR = "Swap2";
    
    public IFixedValueParameter<IntValue> PathsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters["Paths"]; }
    }

    public IFixedValueParameter<BoolValue> BestImprovementParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters["BestImprovement"]; }
    }

    public IValueParameter<IntValue> SeedParameter {
      get { return (IValueParameter<IntValue>)Parameters["Seed"]; }
    }

    public IFixedValueParameter<EnumValue<WalkType>> TypeParameter {
      get { return (IFixedValueParameter<EnumValue<WalkType>>)Parameters["Type"]; }
    }

    public int Paths {
      get { return PathsParameter.Value.Value; }
      set { PathsParameter.Value.Value = value; }
    }

    public bool BestImprovement {
      get { return BestImprovementParameter.Value.Value; }
      set { BestImprovementParameter.Value.Value = value; }
    }

    public int? Seed {
      get { return SeedParameter.Value != null ? SeedParameter.Value.Value : (int?)null; }
      set { SeedParameter.Value = value.HasValue ? new IntValue(value.Value) : null; }
    }

    public WalkType Type {
      get { return TypeParameter.Value.Value; }
      set { TypeParameter.Value.Value = value;}
    }

    [StorableConstructor]
    private QAPDirectedWalk(bool deserializing) : base(deserializing) { }
    private QAPDirectedWalk(QAPDirectedWalk original, Cloner cloner) : base(original, cloner) { }
    public QAPDirectedWalk() {
      characteristics.AddRange(CurveAnalysisResult.AllFeatures.Select(x => new StringValue(x.ToString())));
      characteristics.AddRange(new[] { "AutoCorrelation1", "CorrelationLength", "InformationContent",
        "DensityBasinInformation", "PartialInformationContent", "InformationStability",
        "Diversity", "Regularity", "TotalEntropy", "PeakInformationContent",
        "PeakDensityBasinInformation" }.Select(x => new StringValue(x)));

      Parameters.Add(new FixedValueParameter<IntValue>("Paths", "The number of paths to explore (a path is a set of solutions that connect two randomly chosen solutions).", new IntValue(50)));
      Parameters.Add(new FixedValueParameter<BoolValue>("BestImprovement", "Whether the best of all alternatives should be chosen for each step in the path or just the first improving (least degrading) move should be made.", new BoolValue(true)));
      Parameters.Add(new OptionalValueParameter<IntValue>("Seed", "The seed for the random number generator."));
      Parameters.Add(new FixedValueParameter<EnumValue<WalkType>>("Type", "Type of directed walk to perfom.", new EnumValue<WalkType>(WalkType.RandomRandom)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new QAPDirectedWalk(this, cloner);
    }

    public override bool CanCalculate() {
      return Problem is QuadraticAssignmentProblem;
    }

    private double _evaluatedSolutions;

    public override IEnumerable<IResult> Calculate() {
      _evaluatedSolutions = 0;
      var permutations = CalculateRelinkingPoints();

      var trajectories = Run(permutations).ToList();
      var result = CurveAnalysis<Permutation>.GetCharacteristics(trajectories, (left, right) => (1.0 - HammingSimilarityCalculator.CalculateSimilarity(left, right)) * left.Length);

      #region Calculate Features
      double avgAc1 = 0, avgCorLen = 0, avgIc = 0, avgDbi = 0, avgPic = 0, avgIS = 0, avgDiv = 0,
        avgReg = 0, avgEnt = 0, avgPkIc = 0, avgPkDbi = 0;
      int count = 0;
      foreach (var t in trajectories) {
        var trail = t.Select(x => x.Item2).ToArray();
        if (trail.Length < 4) continue;
        count++;
        double[] acf;
        var len = RuggednessCalculator.CalculateCorrelationLength(trail, out acf);
        avgAc1 += acf[1];
        avgCorLen += len;
        var analysis = new InformationAnalysis(trail);
        avgIc += analysis.InformationContent[0];
        avgDbi += analysis.DensityBasinInformation[0];
        avgPic += analysis.PartialInformationContent[0];
        avgIS += analysis.InformationStability;
        avgDiv += analysis.Diversity;
        avgReg += analysis.Regularity;
        avgEnt += analysis.TotalEntropy[0];
        avgPkIc += analysis.PeakInformationContent.Value;
        avgPkDbi += analysis.PeakDensityBasinInformation.Value;
      }
      avgAc1 /= count;
      avgCorLen /= count;
      avgIc /= count;
      avgDbi /= count;
      avgPic /= count;
      avgIS /= count;
      avgDiv /= count;
      avgReg /= count;
      avgEnt /= count;
      avgPkIc /= count;
      avgPkDbi /= count;
      var isResults = new Dictionary<string, double>() {
        { "AutoCorrelation1", avgAc1 },
        { "CorrelationLength", avgCorLen },
        { "InformationContent", avgIc },
        { "DensityBasinInformation", avgDbi },
        { "PartialInformationContent", avgPic },
        { "InformationStability", avgIS },
        { "Diversity", avgDiv },
        { "Regularity", avgReg },
        { "TotalEntropy", avgEnt },
        { "PeakInformationContent", avgPkIc },
        { "PeakDensityBasinInformation", avgPkDbi }
      };
      #endregion

      foreach (var chara in characteristics.CheckedItems.Select(x => x.Value.Value)) {
        if (Enum.TryParse<CurveAnalysisFeature>(chara, out var caf))
          yield return new Result(Type.ToString() + "-DW." + NBHOOD_STR + "." + chara, new DoubleValue(result.GetValue(caf)));
        else yield return new Result(Type.ToString() + "-DW." + NBHOOD_STR + "." + chara, new DoubleValue(isResults[chara]));
      }
      yield return new Result("EvaluatedSolutions", new IntValue((int)Math.Round(_evaluatedSolutions)));
    }

    private IEnumerable<Permutation> CalculateRelinkingPoints() {
      IRandom random = Seed.HasValue ? new MersenneTwister((uint)Seed.Value) : new MersenneTwister();
      var qap = (QuadraticAssignmentProblem)Problem;
      var pathCount = Paths;

      var perm = new Permutation(PermutationTypes.Absolute, qap.Weights.Rows, random);
      var startLocal = Type == WalkType.LocalGlobal || Type == WalkType.LocalLocal;
      var targetLocal = Type == WalkType.LocalLocal || Type == WalkType.RandomLocal;
      var targetGlobal = Type == WalkType.LocalGlobal || Type == WalkType.RandomGlobal;
      var inverseWalk = Type == WalkType.LocalInverse;

      if (Type == WalkType.LocalInverse) pathCount--; // inverse walks equal one walk per solution
      for (var i = 0; i <= pathCount; i++) {
        var start = i % 2 == 0;
        var target = i % 2 == 1;
        if (inverseWalk || start && startLocal) {
          perm = new Permutation(PermutationTypes.Absolute, qap.Weights.Rows, random);
          var fit = new DoubleValue(QAPEvaluator.Apply(perm, qap.Weights, qap.Distances));
          _evaluatedSolutions++;
          var evalSol = new IntValue(0);
          QAPExhaustiveSwap2LocalImprovement.ImproveFast(perm, qap.Weights, qap.Distances, fit, new IntValue(0), evalSol, qap.Maximization.Value, int.MaxValue, CancellationToken.None);
          _evaluatedSolutions += evalSol.Value;
        } else if (target && (targetLocal || targetGlobal)) {
          if (targetGlobal) {
            perm = qap.BestKnownSolutions.SampleRandom(random);
          } else {
            perm = new Permutation(PermutationTypes.Absolute, qap.Weights.Rows, random);
            var fit = new DoubleValue(QAPEvaluator.Apply(perm, qap.Weights, qap.Distances));
            _evaluatedSolutions++;
            var evalSol = new IntValue(0);
            QAPExhaustiveSwap2LocalImprovement.ImproveFast(perm, qap.Weights, qap.Distances, fit, new IntValue(0), evalSol, qap.Maximization.Value, int.MaxValue, CancellationToken.None);
            _evaluatedSolutions += evalSol.Value;
          }
        } else { // random
          BiasedShuffle(perm, random);
        }
        yield return (Permutation)perm.Clone();
      }
    }

    private IEnumerable<List<Tuple<Permutation, double>>> Run(IEnumerable<Permutation> permutations) {
      if (Type == WalkType.LocalInverse) return RunInverse(permutations);
      return RunRegular(permutations);
    }

    private IEnumerable<List<Tuple<Permutation, double>>> RunInverse(IEnumerable<Permutation> permutations) {
      var qap = (QuadraticAssignmentProblem)Problem;
      var min = qap.LowerBound.Value;
      var max = qap.AverageQuality.Value;
      var bestImprovement = BestImprovement;

      foreach (var start in permutations) {
        var startFitness = QAPEvaluator.Apply(start, qap.Weights, qap.Distances);
        _evaluatedSolutions++;

        // inverse walks are applied to all solutions
        Func<Tuple<Permutation, double>, IEnumerable<Tuple<Permutation, double>>> inverseNeighborFunc = (p) => RestrictedInverseNeighborhood(qap, p.Item1, p.Item2, start);
        var walk = DirectedWalk<Permutation>.WalkRestricted(qap.Maximization.Value, inverseNeighborFunc, start, startFitness, !bestImprovement);
        yield return walk.Select(x => Tuple.Create(x.Item1, (x.Item2 - min) / (max - min))).ToList();
      } // end paths
    }

    private IEnumerable<List<Tuple<Permutation, double>>> RunRegular(IEnumerable<Permutation> permutations) {
      var iter = permutations.GetEnumerator();
      if (!iter.MoveNext()) yield break;
      var bestImprovement = BestImprovement;

      var qap = (QuadraticAssignmentProblem)Problem;
      var min = qap.LowerBound.Value;
      var max = qap.AverageQuality.Value;

      var start = iter.Current;
      
      while (iter.MoveNext()) {
        var startFitness = QAPEvaluator.Apply(start, qap.Weights, qap.Distances);
        _evaluatedSolutions++;
        var end = iter.Current;

        var invSol = new int[start.Length];
        Func<Tuple<Permutation, double>, IEnumerable<Tuple<Permutation, double>>> regularNeighborFunc = (p) => RestrictedNeighborhood(qap, p.Item1, p.Item2, end, invSol);
        var walk = DirectedWalk<Permutation>.WalkRestricted(qap.Maximization.Value, regularNeighborFunc, start, startFitness, !bestImprovement);

        var trail = new List<Tuple<Permutation, double>>();
        foreach (var step in walk) {
          for (var i = 0; i < invSol.Length; i++)
            invSol[step.Item1[i]] = i;
          trail.Add(Tuple.Create(step.Item1, (step.Item2 - min) / (max - min)));
        }
        yield return trail;

        start = end;
      } // end paths
    }

    private IEnumerable<Tuple<Permutation, double>> RestrictedInverseNeighborhood(QuadraticAssignmentProblem qap, Permutation current, double currentFit, Permutation start) {
      var evalSolPerMove = 4.0 / current.Length;
      var N = current.Length;
      for (var index = 0; index < N; index++) {
        if (current[index] != start[index]) continue;
        for (var k = 0; k < N; k++) {
          if (k == index) continue;
          var fit = QAPSwap2MoveEvaluator.Apply(current, new Swap2Move(index, k), qap.Weights, qap.Distances);
          _evaluatedSolutions += evalSolPerMove;
          current.Swap(index, k);
          yield return Tuple.Create(current, currentFit + fit);
          current.Swap(index, k); // reverse
        }
      }
    }

    private IEnumerable<Tuple<Permutation, double>> RestrictedNeighborhood(QuadraticAssignmentProblem qap, Permutation current, double currentFit, Permutation target, int[] inverse) {
      var evalSolPerMove = 4.0 / current.Length;
      for (var index = 0; index < current.Length; index++) {
        if (current[index] == target[index]) continue;
        var k = inverse[target[index]];
        var fit = QAPSwap2MoveEvaluator.Apply(current, new Swap2Move(index, k), qap.Weights, qap.Distances);
        _evaluatedSolutions += evalSolPerMove;
        current.Swap(index, k);
        yield return Tuple.Create(current, currentFit + fit);
        current.Swap(index, k); // reverse
      }
    }

    // permutation must be strictly different in every position
    private static void BiasedShuffle(Permutation p, IRandom random) {
      for (var i = p.Length - 1; i > 0; i--) {
        // Swap element "i" with a random earlier element (excluding itself)
        var swapIndex = random.Next(i);
        var h = p[swapIndex];
        p[swapIndex] = p[i];
        p[i] = h;
      }
    }
  }
}
